import copy
import os
from typing import Dict, List, Optional

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from stl import mesh
from tqdm import tqdm

from pyskindose.plotting.create_ploty_ijk_indices import _create_plotly_ijk_indices_for_cuboid_objects

from .settings_pyskindose import PhantomDimensions

# valid phantom types
VALID_PHANTOM_MODELS = ["plane", "cylinder", "human", "table", "pad"]


class Phantom:
    """Create and handle phantoms for patient, support table and pad.

    This class creates a phatom of any of the types specified in
    VALID_PHANTOM_MODELS (plane, cylinder or human to represent the patient,
    as well as patient support table and pad). The patient phantoms consists of
    a number of skin cells where the skin dose can be calculated.

    Attributes
    ----------
    phantom_model : str
        Type of phantom, i.e. "plane", "cylinder", "human", "table" or "pad"
    r : np.array
        n*3 array where n are the number of phantom skin cells. Each row
        contains the xyz coordinate of one of the phantom skin cells
    ijk : np.array
        A matrix containing vertex indices. This is required in order to
        plot the phantom using plotly Mesh3D. For more info, see "i", "j", and
        "k" at https://plot.ly/python/reference/#mesh3d
    dose : np.array
        An empty 1d array to store skin dose calculation for each of the n
        phantom cells. Only for patient phantom types (plane, cylinder, human)
    n : np.array
        normal vectors to each of the n phantom skin cells. (only for 3D
        patient phantoms, i.e. "cylinder" and "human")
    r_ref : np.array
        Empty array to store of reference position of the phantom cells after
        the phantom has been aligned in the geometry with the
        position_patient_phantom_on_table function in geom_calc.py
    table_length : float
        length of patient support table. The is needed for all phantom object
        to select correct rotation origin for At1, At2, and At3.

    Methods
    -------
    rotate(rotation)
        Rotating the phantom about any of the x, y, or z axis
    translate(dr)
        Translates the phantom along the x, y or z direction
    save_position
        Saves the reference position after the phantom has been properly
        positioned in the irradiation geometry. This method is called in the
        position_patient_phantom_on_table function
    position(data_norm)
        Positions the phantom from reference position to actual position
        according to the table displacement info in data_norm
    """

    def __init__(self, phantom_model: str, phantom_dim: PhantomDimensions, human_mesh: Optional[str] = None):
        """Create the phantom of choice.

        Parameters
        ----------
        phantom_model : str
            Type of phantom to create. Valid selections are 'plane',
            'cylinder', 'human', "table" an "pad".
        phantom_dim : PhantomDimensions
            instance of class PhantomDimensions containing dimensions for
            all phantoms models except human phantoms: Length, width, radius,
            thickness etc.
        human_mesh : str, optional
            Choose which human mesh phantom to use. Valid selection are names
            of the *.stl-files in the phantom_data folder (The default is none.

        Raises
        ------
        ValueError
            Raises value error if unsupported phantom type are selected,
            or if phantom_model='human' selected, without specifying
            human_mesh

        """
        self.phantom_model = phantom_model.lower()
        # Raise error if invalid phantom model selected
        if self.phantom_model not in VALID_PHANTOM_MODELS:
            raise ValueError(f"Unknown phantom model selected. Valid type:" f"{'.'.join(VALID_PHANTOM_MODELS)}")

        self.r_ref: np.array

        # Save table length for all phantom in order to choose correct rotation
        # origin when applying At1, At2, and At3
        self.table_length = phantom_dim.table_length

        # creates a plane phantom (2D grid)
        if phantom_model == "plane":

            # Use a dense grid if specified by user
            if phantom_dim.plane_resolution.lower() == "dense":
                res_length = res_width = 2.0

            elif phantom_dim.plane_resolution.lower() == "sparse":
                res_length = res_width = 1.0

            # Linearly spaced points along the longitudinal direction
            x = np.linspace(
                -phantom_dim.plane_width / 2, +phantom_dim.plane_width / 2, int(res_width * phantom_dim.plane_width + 1)
            )
            # Linearly spaced points along the lateral direction
            z = np.linspace(0, -phantom_dim.plane_length, int(res_length * phantom_dim.plane_length))

            # Create phantom in form of rectangular grid
            x_plane, z_plane = np.meshgrid(x, z)

            t = phantom_dim.plane_width

            # Create index vectors for plotly mesh3d plotting
            i2: List[int] = []
            i1 = j1 = k1 = i2

            for i in range(len(x) - 1):
                for j in range(len(z) - 1):
                    i1 = i1 + [j * len(x) + i]
                    j1 = j1 + [j * len(x) + i + 1]
                    k1 = k1 + [j * len(x) + i + len(x)]
                    i2 = i2 + [j * len(x) + i + len(x) + 1]

            self.r = np.column_stack((x_plane.ravel(), np.zeros(len(x_plane.ravel())), z_plane.ravel()))

            self.ijk = np.column_stack((i1 + i2, j1 + k1, k1 + j1))
            self.dose = np.zeros(len(self.r))

        # creates a cylinder phantom (elliptic)
        elif phantom_model == "cylinder":

            # Use a dense grid if specified by user
            if phantom_dim.cylinder_resolution.lower() == "dense":
                res_length = 4
                res_width = 0.05

            elif phantom_dim.cylinder_resolution.lower() == "sparse":
                res_length = 1.0
                res_width = 0.1

            # Creates linearly spaced points along an ellipse
            #  in the lateral direction
            t = np.arange(0 * np.pi, 2 * np.pi, res_width)
            x = (phantom_dim.cylinder_radii_a * np.cos(t)).tolist()
            y = (phantom_dim.cylinder_radii_b * np.sin(t)).tolist()

            # calculate normal vectors of a cylinder (pointing outwards)
            nx = np.cos(t) / (np.sqrt(np.square(np.cos(t) + 4 * np.square(np.sin(t)))))

            nz = np.zeros(len(t))

            ny = 2 * np.sin(t) / (np.sqrt(np.square(np.cos(t) + 4 * np.square(np.sin(t)))))

            nx = nx.tolist()
            ny = ny.tolist()
            nz = nz.tolist()

            n = [[nx[ind], ny[ind], nz[ind]] for ind in range(len(t))]

            # Store the  coordinates of the cylinder phantom
            output: Dict = dict(n=[], x=[], y=[], z=[])

            # Extend the ellipse to span the entire length of the phantom,
            # thus creating an elliptic cylinder
            for index in range(0, int(res_length) * (phantom_dim.cylinder_length + 2), 1):

                output["x"] = output["x"] + x
                output["z"] = output["z"] + [-1 / res_length * index] * len(x)
                output["y"] = output["y"] + y
                output["n"] = output["n"] + n

            # Create index vectors for plotly mesh3d plotting
            i1 = list(range(0, len(output["x"]) - len(t)))
            j1 = list(range(1, len(output["x"]) - len(t) + 1))
            k1 = list(range(len(t), len(output["x"])))
            i2 = list(range(0, len(output["x"]) - len(t)))
            k2 = list(range(len(t) - 1, len(output["x"]) - 1))
            j2 = list(range(len(t), len(output["x"])))

            for i in range(len(output["y"])):
                output["y"][i] -= phantom_dim.cylinder_radii_b

            self.r = np.column_stack((output["x"], output["y"], output["z"]))
            self.ijk = np.column_stack((i1 + i2, j1 + j2, k1 + k2))
            self.dose = np.zeros(len(self.r))
            self.n = np.asarray(output["n"])

        # creates a human phantom
        elif phantom_model == "human":

            if human_mesh is None:
                raise ValueError("Human model needs to be specified for" 'phantom_model = "human"')

            # load selected phantom model from binary .stl file
            phantom_path = os.path.join(os.path.dirname(__file__), "phantom_data", f"{human_mesh}.stl")
            phantom_mesh = mesh.Mesh.from_file(phantom_path)

            r = phantom_mesh.vectors
            n = phantom_mesh.normals

            self.r = np.asarray([el for el_list in r for el in el_list])
            self.n = np.asarray([x for pair in zip(n, n, n) for x in pair])

            # Create index vectors for plotly mesh3d plotting
            self.ijk = np.column_stack(
                (np.arange(0, len(self.r) - 3, 3), np.arange(1, len(self.r) - 2, 3), np.arange(2, len(self.r) - 1, 3))
            )
            self.dose = np.zeros(len(self.r))

        # Creates the vertices of the patient support table
        elif phantom_model == "table":
            # Longitudinal position of the the vertices
            x_tab = [index * phantom_dim.table_width for index in [+0.5, +0.5, -0.5, -0.5, +0.5, +0.5, -0.5, -0.5]]

            # Vertical position of the vertices
            y_tab = [index * phantom_dim.table_thickness for index in [0, 0, 0, 0, +1, +1, +1, +1]]

            # Lateral position of the vertices
            z_tab = [index * phantom_dim.table_length for index in [0, -1, -1, 0, 0, -1, -1, 0]]

            # Create index vectors for plotly mesh3d plotting
            i_tab, j_tab, k_tab = _create_plotly_ijk_indices_for_cuboid_objects()

            self.r = np.column_stack((x_tab, y_tab, z_tab))
            self.ijk = np.column_stack((i_tab, j_tab, k_tab))

        # Creates the vertices of the patient support table
        elif phantom_model == "pad":

            # Longitudinal position of the the vertices
            x_pad = [index * phantom_dim.pad_width for index in [+0.5, +0.5, -0.5, -0.5, +0.5, +0.5, -0.5, -0.5]]

            # Vertical position of the vertices
            y_pad = [index * phantom_dim.pad_thickness for index in [0, 0, 0, 0, -1, -1, -1, -1]]

            # Lateral position of the the vertices
            z_pad = [index * phantom_dim.pad_length for index in [0, -1, -1, 0, 0, -1, -1, 0]]

            # Create index vectors for plotly mesh3d plotting
            i_pad, j_pad, k_pad = _create_plotly_ijk_indices_for_cuboid_objects()

            self.r = np.column_stack((x_pad, y_pad, z_pad))
            self.ijk = np.column_stack((i_pad, j_pad, k_pad))

    def rotate(self, angles: List[int]) -> None:
        """Rotate the phantom about the angles specified in rotation.

        Parameters
        ----------
        angles: List[int]
            list of angles in degrees the phantom should be rotated about,
            given as [x_rot: <int>, y_rot: <int>, z_rot: <int>]. E.g.
            rotation = [0, 90, 0] will rotate the phantom 90 degrees about the
            y-axis.

        """
        # convert degrees to radians
        angles = np.deg2rad(angles)

        x_rot = angles[0]
        y_rot = angles[1]
        z_rot = angles[2]

        # Define rotation matricies about the x, y and z axis
        Rx = np.array([[+1, +0, +0], [+0, +np.cos(x_rot), -np.sin(x_rot)], [+0, +np.sin(x_rot), +np.cos(x_rot)]])
        Ry = np.array([[+np.cos(y_rot), +0, +np.sin(y_rot)], [+0, +1, +0], [-np.sin(y_rot), +0, +np.cos(y_rot)]])
        Rz = np.array([[+np.cos(z_rot), -np.sin(z_rot), +0], [+np.sin(z_rot), +np.cos(z_rot), +0], [+0, +0, +1]])

        # Rotate position vectors to the phantom cells

        self.r = np.matmul(Rx, np.matmul(Ry, np.matmul(Rz, self.r.T))).T

        if self.phantom_model in ["cylinder", "human"]:

            self.n = np.matmul(Rx, np.matmul(Ry, np.matmul(Rz, self.n.T))).T

    def translate(self, dr: List[int]) -> None:
        """Translate the phantom in the x, y or z direction.

        Parameters
        ----------
        dr : List[int]
            list of distances the phantom should be translated, given in cm.
            Specified as dr = [dx: <int>, dy: <int>, dz: <int>]. E.g.
            dr = [0, 0, 10] will translate the phantom 10 cm in the z direction

        """
        self.r[:, 0] += dr[0]
        self.r[:, 1] += dr[1]
        self.r[:, 2] += dr[2]

    def save_position(self) -> None:
        """Store a reference position of the phantom.

        This function is supposed to be used to store the patient fixation
        conducted in the function position_patient_phantom_on_table

        """
        r_ref = copy.copy(self.r)
        self.r_ref = r_ref

    def position(self, data_norm: pd.DataFrame, event: int) -> None:
        """Position the phantom for a event by adding RDSR table displacement.

        Positions the phantom from reference position to actual position
        according to the table displacement info in data_norm.

        Parameters
        ----------
        data_norm : pd.DataFrame
            Table containing dicom RDSR information from each irradiation event
            See rdsr_normalizer.py for more information.
        event : int
            Irradiation event index

        """
        self.r = copy.copy(self.r_ref)

        # Fetch rotation angles of the patient, table, and pad

        # Table Horizontal Rotation Angle (At1)
        # i.e. rotation of the table about the positive y axis (VERT),
        # with rotation axis in the center of the table.
        at1 = np.deg2rad(data_norm["At1"][event])
        # Table Head Tilt Angle (At2)
        # i.e. rotation of the table about the positive x axis (LON)
        # with rotation axis in the center of the table.
        at2 = np.deg2rad(data_norm["At2"][event])
        # Table Cradle Tilt Angle (At3)
        # i.e. rotation of the table about the z axis (LAT)
        at3 = np.deg2rad(data_norm["At3"][event])

        # displace phantom to table rotation center
        self.r[:, 2] += self.table_length / 2

        # calculate rotation about x axis
        angle = at2
        Rx = np.array([[+1, +0, +0], [+0, +np.cos(angle), -np.sin(angle)], [+0, +np.sin(angle), +np.cos(angle)]])

        # calculate rotation about y axis
        angle = at1
        Ry = np.array([[+np.cos(angle), +0, +np.sin(angle)], [+0, +1, +0], [-np.sin(angle), +0, +np.cos(angle)]])

        # calculate rotation about z axis
        angle = at3
        Rz = np.array([[+np.cos(angle), -np.sin(angle), +0], [+np.sin(angle), +np.cos(angle), +0], [+0, +0, +1]])

        # Apply table rotation
        self.r = np.matmul(Rz, np.matmul(Ry, np.matmul(Rx, self.r.T))).T

        # Replace phantom back to starting position
        self.r[:, 2] -= self.table_length / 2

        # Apply phantom translation
        t = np.array([data_norm.Tx[event], data_norm.Ty[event], data_norm.Tz[event]])

        self.r = self.r + t



from typing import Any, Dict, Union

import pandas as pd

from pyskindose import constants as c
from pyskindose.calculate_dose.calculate_dose import calculate_dose
from pyskindose.phantom_class import Phantom
from pyskindose.plotting.create_dose_map_plot import create_dose_map_plot
from pyskindose.plotting.create_geometry_plot import create_geometry_plot
from pyskindose.settings_pyskindose import PyskindoseSettings, initialize_settings


def analyze_data(
    normalized_data: pd.DataFrame,
    settings: Union[str, dict, PyskindoseSettings],
) -> Dict[str, Any]:
    """Analyze data och settings, and runs PySkinDose in desired mode.

    Parameters
    ----------
    normalized_data : pd.DataFrame
        RDSR data, normalized for compliance with PySkinDose.
    settings : Union[str, dict, PyskindoseSettings]
        Settings class for PySkinDose

    Returns
    -------
    Dict[str, Any]
        output dictionary containing calculation specifics such as dose map, correction
        factors, etc.

    """
    settings = initialize_settings(settings)

    # create table, pad and patient phantoms.
    table = Phantom(phantom_model=c.PHANTOM_MODEL_TABLE, phantom_dim=settings.phantom.dimension)

    pad = Phantom(phantom_model=c.PHANTOM_MODEL_PAD, phantom_dim=settings.phantom.dimension)

    create_geometry_plot(normalized_data=normalized_data, table=table, pad=pad, settings=settings)

    patient, output = calculate_dose(normalized_data=normalized_data, settings=settings, table=table, pad=pad)

    # Fetch dose_map if calculate_dose has been executed
    dose_map = output[c.OUTPUT_KEY_DOSE_MAP] if settings.mode == c.MODE_CALCULATE_DOSE else None

    create_dose_map_plot(patient=patient, settings=settings, dose_map=dose_map)

    return output



import json
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd

from .constants import (
    KEY_NORMALIZATION_FILTER_SIZE_ALUMINUM,
    KEY_NORMALIZATION_FILTER_SIZE_COPPER,
    KEY_RDSR_FILTER_MATERIAL,
    KEY_RDSR_FILTER_MATERIAL_ALUMINUM,
    KEY_RDSR_FILTER_MATERIAL_COPPER,
    KEY_RDSR_FILTER_MAX,
    KEY_RDSR_FILTER_MIN,
    KEY_NORMALIZATION_DISTANCE_SOURCE_DETECTOR,
    KEY_NORMALIZATION_DISTANCE_SOURCE_ISOCENTER,
    KEY_NORMALIZATION_DISTANCE_ISOCENTER_DETECTOR,
    KEY_NORMALIZATION_DISTANCE_SOURCE_IRP,
    KEY_NORMALIZATION_MODEL_NAME,
    KEY_NORMALIZATION_ACQUISITION_TYPE,
    KEY_NORMALIZATION_ACQUISITION_PLANE,
    KEY_RDSR_DISTANCE_SOURCE_DETECTOR,
)
from .geom_calc import calculate_field_size
from .settings_normalization import NormalizationSettings


def rdsr_normalizer(
    data_parsed: pd.DataFrame, normalization_settings: Optional[Union[str, dict]] = None
) -> pd.DataFrame:
    """Normalize RDSR data for PySkinDose compliance.

    Parameters
    ----------
    data_parsed : pd.DataFrame
        Parsed RDSR data from all irradiation events in the RDSR input file,
        i.e. output of function rdsr_parser

    Returns
    -------
    data_norm : pd.DataFrame
        Dataframe with the following columns:
    model : str
        device model, e.g. 'AXIOM-artis
    DSD : float
        Distance Source to detector (DSD) in cm.
    DSI : float
        Distance Source to Isocenter (DSI) in cm.
    DID : float
        Distance Isocenter to Detector (DID) in cm.
    DSIRP : float
        Distance Source to intercentional reference point (DSIRP) in cm.
    acquisition_type : str
        Type of irradiation event, i.e. 'fluoroscopy, or stationary
        acquisition.
    acquisition_plane : str
        plane used for image acquisition. Either 'single plane', 'plane a', or
        plane b.
    Tx : float
        Table offset in x-direction (longitudinal direction) from the machine
        isocenter. At Tx = 0, the patient support table is centered about the
        isocenter x-axis. With patient lying in head-first supine position
        (default settings), Tx increases in patient left lateral direction.
    Ty : float
        Table offset in y-direction (vertical direction) from the machine
        isocenter. At Ty = 0, the patient support table is centered about the
        isocenter y-axis. Ty is increasing downwards, i.e. along the force
        of gravity.
    Tz : float
        Table offset in z-direction (lateral direction) from the machine
        isocenter. At Tz = 0, the head end of the patient support tables are
        located at the zero coordinate of the z-axis. With patient lying in
        head-first supine position (default settings), Tz increases in the
        patients cranial direction.

        .. image:: user/figures/table/table_translate.svg

    At1 : int
        Rotation angle of the patient support table about the isocenter
        y-axis. The center of rotation is located at the centerpoint
        of the table. Positive direction is determined by the right-hand rule
        for curve orientation about the positive isocenter y-axis.

        .. image:: user/figures/table/table_at1.svg

    At2 : int
        Tilt angel of the patient support table about the isocenter x-axis.
        The center of the tilt is located at the center of the table, with
        positive direction determined by the right-hand rule for curve
        orientations  about the positiove isocenter x-axis.

        .. image:: user/figures/table/table_at2.svg

    At3 : int
        Cradle angle of the patient support table about the isocenter
        z-axis. The center of rotation is located at the centerpoint
        of the table. Positive direction is determined by the right-hand rule
        for curve orientation about the positive isocenter z-axis.

        .. image:: user/figures/table/table_at3.svg

    Ap1 : int
        Rotation angle of the X-ray source about the isocenter z-axis. Positive
        direction is determined by the right-hand rule for curve orientation
        about the positive isocenter z-axis.

        .. image:: user/figures/beam/beam_ap1.svg

    Ap2 : int
        Rotation angle of the X-ray source about the isocenter x-axis. Positive
        direction is determined by the right-hand rule for curve orientation
        about the positive isocenter x-axis.

        .. image:: user/figures/beam/beam_ap2.svg

    Ap3 : int
        Rotation angle of the X-ray detector about the isocenter y-axis.
        Positive direction is determined by the right-hand rule for curve
        orientation about the positive isocenter y-axis.


    DSL : float
        Detector Side Length (DSL) in cm.
    FS_lat : float
        Side length of the X-ray field in the lateral direction at the image
        receptor plane.
    FS_long : float
        Side length of the X-ray field in the longitudinal direction at the
        image receptor plane.
    kVp : float
        Tube voltage in kV
    K_IRP : float
        IRP air kerma at the Interventional Reference Point (IRP).
    filter_thickness_Cu : float
        Copper X-ray filter thickness in mm.
    filter_thickness_Al : float
        Aluminum X-ray filter thickness in mm.
    """
    data_norm = pd.DataFrame()

    norm = _load_normalization_settings(data_parsed=data_parsed, norm_settings=normalization_settings)

    for append_normalization in [
        _normalize_machine_parameters,
        _normalize_table_parameters,
        _normalize_xray_filter_materials,
        _normalize_beam_parameters,
    ]:
        data_norm = append_normalization(data_parsed=data_parsed, data_norm=data_norm, norm=norm)

    return data_norm


def _load_normalization_settings(
    data_parsed: pd.DataFrame, norm_settings: Optional[Union[str, dict]] = None
) -> NormalizationSettings:

    if norm_settings is None:

        normalization_settings_path = Path(__file__).parent / "normalization_settings.json"

        with normalization_settings_path.open("r") as json_file:
            norm_settings = json.load(json_file)

    if isinstance(norm_settings, str):
        norm_settings = json.load(json_file)

    return NormalizationSettings(normalization_settings=norm_settings, data_parsed=data_parsed)


def _normalize_machine_parameters(
    data_parsed: pd.DataFrame, data_norm: pd.DataFrame, norm: NormalizationSettings
) -> pd.DataFrame:

    data_norm[KEY_NORMALIZATION_MODEL_NAME] = data_parsed.ManufacturerModelName

    # Find indices of nans in DistanceSourcetoDetector
    if "nan" in str(data_parsed[KEY_RDSR_DISTANCE_SOURCE_DETECTOR]).lower():
        nan_indices = data_parsed.index[data_parsed[KEY_RDSR_DISTANCE_SOURCE_DETECTOR].apply(np.isnan)]
        # Replace those nans with the corresponding value in
        # FinalDistanceSourcetoDetector
        data_parsed.DistanceSourcetoDetector_mm = data_parsed.DistanceSourcetoDetector_mm.fillna(
            data_parsed.FinalDistanceSourcetoDetector_mm[nan_indices]
        )

    data_norm[KEY_NORMALIZATION_DISTANCE_SOURCE_DETECTOR] = data_parsed.DistanceSourcetoDetector_mm / 10
    data_norm[KEY_NORMALIZATION_DISTANCE_SOURCE_ISOCENTER] = data_parsed.DistanceSourcetoIsocenter_mm / 10
    data_norm[KEY_NORMALIZATION_DISTANCE_ISOCENTER_DETECTOR] = data_norm.DSD - data_norm.DSI
    data_norm[KEY_NORMALIZATION_DISTANCE_SOURCE_IRP] = data_norm.DSI - 15
    data_norm[KEY_NORMALIZATION_ACQUISITION_TYPE] = data_parsed.IrradiationEventType
    data_norm[KEY_NORMALIZATION_ACQUISITION_PLANE] = data_parsed.AcquisitionPlane

    return data_norm


def _normalize_table_parameters(
    data_parsed: pd.DataFrame, data_norm: pd.DataFrame, norm: NormalizationSettings
) -> pd.DataFrame:

    # Table translations
    data_norm["Tx"] = norm.trans_offset.x + norm.trans_dir.x * data_parsed.TableLongitudinalPosition_mm / 10

    data_norm["Ty"] = norm.trans_offset.y + norm.trans_dir.y * data_parsed.TableHeightPosition_mm / 10

    data_norm["Tz"] = norm.trans_offset.z + norm.trans_dir.z * data_parsed.TableLateralPosition_mm / 10

    # Table rotations
    data_norm["At1"] = norm.rot_dir.At1 * [0] * len(data_norm)
    data_norm["At2"] = norm.rot_dir.At2 * [0] * len(data_norm)
    data_norm["At3"] = norm.rot_dir.At3 * [0] * len(data_norm)

    return data_norm


def _normalize_xray_filter_materials(
    data_parsed: pd.DataFrame, data_norm: pd.DataFrame, norm: NormalizationSettings
) -> pd.DataFrame:
    # parse filter material and thickness

    # Load filter min and max, and fill all NANs with zeros
    for key in [KEY_RDSR_FILTER_MIN, KEY_RDSR_FILTER_MAX]:
        data_parsed[key] = data_parsed[key].fillna(0.0)

    # Add columns for filter materials in data_norm, and initialize to zero
    for key in [
        KEY_NORMALIZATION_FILTER_SIZE_COPPER,
        KEY_NORMALIZATION_FILTER_SIZE_ALUMINUM,
    ]:
        data_norm[key] = 0.0

    # for each irradiation event
    for event_index in range(len(data_parsed)):

        # fetch filter materials from data_parsed
        event_filter_materials = data_parsed[KEY_RDSR_FILTER_MATERIAL][event_index]

        # fetch filter min and max thicknesses
        event_filter_minmax = np.array(
            [data_parsed[KEY_RDSR_FILTER_MIN][event_index], data_parsed[KEY_RDSR_FILTER_MAX][event_index],]
        )

        # calculate filter mean thicknesses
        event_filter_means = np.mean(event_filter_minmax, axis=0)

        # Make list of event_filter_means (required if only 1 filter material (e.g. Axiom Artis))
        if isinstance(event_filter_materials, str):
            event_filter_means = [event_filter_means]

        # if copper filter in use
        if KEY_RDSR_FILTER_MATERIAL_COPPER in event_filter_materials:
            # append copper filtration to normalized data
            data_norm.loc[event_index, KEY_NORMALIZATION_FILTER_SIZE_COPPER] = event_filter_means[
                event_filter_materials.index(KEY_RDSR_FILTER_MATERIAL_COPPER)
            ]
        # if aluminum filter in use
        if KEY_RDSR_FILTER_MATERIAL_ALUMINUM in event_filter_materials:
            # append aluminum filtration to normalized data
            data_norm.loc[event_index, KEY_NORMALIZATION_FILTER_SIZE_ALUMINUM] = event_filter_means[
                event_filter_materials.index(KEY_RDSR_FILTER_MATERIAL_ALUMINUM)
            ]

    return data_norm


def _normalize_beam_parameters(
    data_parsed: pd.DataFrame, data_norm: pd.DataFrame, norm: NormalizationSettings
) -> pd.DataFrame:

    # beam angulation
    data_norm["Ap1"] = norm.rot_dir.Ap1 * data_parsed.PositionerPrimaryAngle_deg
    data_norm["Ap2"] = norm.rot_dir.Ap2 * data_parsed.PositionerSecondaryAngle_deg
    # temp set to zero
    data_norm["Ap3"] = norm.rot_dir.Ap3 * [0] * len(data_norm)

    # detector side length
    data_norm["DSL"] = norm.detector_side_length

    FS_lat, FS_long = calculate_field_size(
        field_size_mode=norm.field_size_mode, data_parsed=data_parsed, data_norm=data_norm,
    )

    data_norm["FS_lat"] = FS_lat
    data_norm["FS_long"] = FS_long

    data_norm["kVp"] = data_parsed.KVP_kV
    data_norm["K_IRP"] = data_parsed.DoseRP_Gy * 1000

    return data_norm



import argparse
import logging
import os
from typing import Optional, Union

import pandas as pd
import pydicom

from pyskindose.analyze_data import analyze_data
from pyskindose.dev_data import DEVELOPMENT_PARAMETERS
from pyskindose.rdsr_normalizer import rdsr_normalizer
from pyskindose.rdsr_parser import rdsr_parser
from pyskindose.settings_pyskindose import PyskindoseSettings, initialize_settings

logger = logging.getLogger(__name__)


def main(file_path: Optional[str] = None, settings: Union[str, dict, PyskindoseSettings] = None):
    """Run PySkinDose.

    Copy settings_examples.json and save it as settings.json.
    Set all you parameters in this file. For debugging and developement,
    the PARAM_dev settings dictionary can be used by calling
    main(settings=PARAM_DEV).

    See settings.py for a description of all the parameters. Please visit
    https://github.com/rvbCMTS/PySkinDose for info on how to run
    PySkinDose.

    Parameters
    ----------
    file_path : str, optional
        file path to RDSR file or preparsed RDSR data in .json format
    settings : Union[str, dict, PyskindoseSettings], optional
        Setting file in either dict, json string format, or as a
        PyskindoseSettings object. By default, settings_examples.json is
        enabled.

    """
    settings = _parse_settings_to_settings_class(settings=settings)

    data_norm = _read_and_normalise_rdsr_data(rdsr_filepath=file_path, settings=settings)

    _ = analyze_data(normalized_data=data_norm, settings=settings)


def _parse_settings_to_settings_class(settings: Optional[str] = None):
    try:
        return initialize_settings(settings)
    except ValueError:
        logger.debug("Tried initializing settings without any settings")

    settings_path = os.path.join(os.path.dirname(__file__), "settings.json")

    if not os.path.exists(settings_path):
        logger.warning("Settings path not specified. Using example settings.")
        settings_path = os.path.join(os.path.dirname(__file__), "settings_example.json")

    with open(settings_path, "r") as fp:
        output = fp.read()

    return PyskindoseSettings(output)


def _read_and_normalise_rdsr_data(rdsr_filepath: str, settings: PyskindoseSettings):

    if not rdsr_filepath:
        rdsr_filepath = os.path.join(os.path.dirname(__file__), "example_data", "RDSR", settings.rdsr_filename)

    logger.debug(rdsr_filepath)

    "If provided, load preparsed rdsr data in .json format"
    if ".json" in rdsr_filepath:
        return pd.read_json(rdsr_filepath)

    # else load RDSR data with pydicom
    data_raw = pydicom.dcmread(rdsr_filepath)

    # parse RDSR data from raw .dicom file
    data_parsed = rdsr_parser(data_raw)

    # normalized rdsr for compliance with PySkinDose
    normalized_data = rdsr_normalizer(data_parsed)

    return normalized_data


if __name__ == "__main__":

    DESCRIPTION = (
        "PySkinDose is a Python version 3.7 based program for patient peak"
        " skin dose (PSD) estimations from fluoroscopic procedures in"
        " interventional radiology."
    )

    PARSER = argparse.ArgumentParser(description=DESCRIPTION)
    PARSER.add_argument("--file-path", help="Path to RDSR DICOM file")
    ARGS = PARSER.parse_args()

    main(file_path=ARGS.file_path, settings=DEVELOPMENT_PARAMETERS)



import json
from typing import Union

from pyskindose.constants import (
    KEY_PARAM_ESTIMATE_K_TAB,
    KEY_PARAM_HUMAN_MESH,
    KEY_PARAM_K_TAB_VAL,
    KEY_PARAM_MODE,
    KEY_PARAM_PHANTOM_MODEL,
    KEY_PARAM_RDSR_FILENAME,
    OFFSET_LATERAL_KEY,
    OFFSET_LONGITUDINAL_KEY,
    OFFSET_VERTICAL_KEY,
)


class PyskindoseSettings:
    """A class to store all settings required to run PySkinDose.

    Attributes
    ----------
    mode : str
        Select which mode to execute PySkinDose with. There are three
        different modes:

        mode = "calculate_dose" calculates the skindose from the RDSR data and
        presents the result in a skindose map.

        mode = "plot_setup" plots the geometry (patient, table, pad and beam
        in starting position, i.e., before any RDSR data has been added.) This
        is useful for debugging and when manually fixating the patient phantom
        with the function "position_patient_phantom_on_table".

        mode = "plot_event" plots the geometry for a specific irradiation event
        with index = event.

        mode = "plot_procedure" plots geometry of the entire sequence of RDSR
        events provided in the RDSR file. The patient phantom is omitted for
        calculation speed in human phantom is used.

    rdsr_filename : str
        filename of the RDSR file, without the .dcm file ending.
    estimate_k_tab : bool
        Wheter k_tab should be approximated or not. You this if have not
        conducted table attenatuion measurements.
    k_tab_val : float
        Value of k_tab, in range 0.0 -> 1.0.
    phantom : PhantomSettings
        Instance of class PhantomSettings containing all phantom related
        settings.
    plot : Plotsettings
        Instace of class Plottsettings containing all plot related settings

    """

    def __init__(self, settings: Union[str, dict]):
        """Initialize settings class.

        Parameters
        ----------
        settings : Union[str, dict]
            settings : Union[str, dict]
            Either a .json string or a dictionary containing all the settings
            parameters required to run PySkinDose. See setting_example.json
            in /settings/ for example.

        """
        if isinstance(settings, str):
            tmp = json.loads(settings)
        else:
            tmp = settings

        self.mode = tmp[KEY_PARAM_MODE]
        self.k_tab_val = tmp[KEY_PARAM_K_TAB_VAL]
        self.rdsr_filename = tmp[KEY_PARAM_RDSR_FILENAME]
        self.estimate_k_tab = tmp[KEY_PARAM_ESTIMATE_K_TAB]

        self.phantom = PhantomSettings(ptm_dim=tmp["phantom"])
        self.plot = Plotsettings(plt_dict=tmp["plot"])

    def print_parameters(self, return_as_string: bool = False):
        """Print entire parameter class to terminal.

        Parameters
        ----------
        return_as_string : bool, optional
            Return the print statement as a string, instead of printing it
            to the terminal. The default is False.

        """
        self.phantom.patient_offset.update_attrs_str()
        self.phantom.dimension.update_attrs_str()
        self.phantom.update_attrs_str()
        self.plot.update_attrs_str()

        main_attrs_str = create_attrs_str(attrs_parent=self, object_name="general", indent_level=0)

        if return_as_string:
            return main_attrs_str

        return print(main_attrs_str)


class PhantomSettings:
    """A class for setting all the phantom related settings required.

    Attributes
    ----------
    model : str
        Select which model to represent the skin surface for skindose
        calculations. Valid selections: "plane" (2D planar surface),
        "cylinder" (cylinder with elliptical cross section) or "human" (phantom
        in the shape of a human, created with MakeHuman.)
    human_mesh: str
        Select which MakeHuman phantom to represent the patient when
        model = "human" is selected. Valid selections: Any of the .stl files
        in the folder phantom_data. Enter as a string without the .stl file
        ending.
    patient_offset : PhantomOffset
        Instance of class PhantomOffset containing patient - table isocenter
        offset.
    patient_orientation : str
        patient orientation on table. Choose between 'head_first_supine' and
        'feet_first_supine'.
    dimension : PhantomDimensions
        Instance of class PhantomDimensions containing all dimensions required
        to create any of the mathematical phantoms, which is all but human.

    """

    def __init__(self, ptm_dim: dict):
        """Initialize phantom settings class.

        Parameters
        ----------
        ptm_dim : PhantomDimensions
            Instance of class PhantomDimensions containing all dimensions for
            the mathematical phantoms. See attributes of PhantomDimensions.

        """
        self.model = ptm_dim[KEY_PARAM_PHANTOM_MODEL]
        self.human_mesh = ptm_dim[KEY_PARAM_HUMAN_MESH]
        self.patient_orientation = ptm_dim["patient_orientation"]
        self.patient_offset = PatientOffset(offset=ptm_dim["patient_offset"])
        self.dimension = PhantomDimensions(ptm_dim=ptm_dim["dimension"])

        self.attrs_str = create_attrs_str(attrs_parent=self, object_name="phantom", indent_level=0)
        return

    def update_attrs_str(self):
        self.attrs_str = create_attrs_str(attrs_parent=self, object_name="phantom", indent_level=0)


class PhantomDimensions:
    """A class for setting the phantom dimensions for mathematical phantoms.

    Attributes
    ----------
    plane_length : int
        Lenth of plane phantom.
    plane_width : int
        Width of plane phantom.
    plane_resolution: str
        Select either 'sparse' or 'dense' resolution of the skin cell grid
        on the surface of the plane phantom. Note: dense is more computational
        expensive.
    cylinder_length : int
        Length of cylider phantom.
    cylinder_radii_a : float
        First radii of the cylindrical cross section of the cylindrical
        phantom, which lies in the "width" direction.
    cylinder_radii_b : float
        Second radii of the cylindrical cross section of the cylindrical
        phantom, which lies in the "thickness" direction. radii a should
        be greater than radii b.
    cylinder_resolution: str
        Select either 'sparse' or 'dense' resolution of the skin cell grid
        on the surface of the elliptical cylinder. Note: dense is more
        computational expensive.
    table_thickness : int
        Thickness of the support table phantom.
    table_length : int
        Length of the support table phantom.
    table_width : int
        Width of the support table phantom.
    pad_thickness : int
        Thickness of the patient support table phantom.
    pad_width : int
        Width of the patient support table phantom.
    pad_length : int
        Length of the patient support table phantom.

    """

    def __init__(self, ptm_dim: dict):
        """Initialize phantom dimension class.

        Parameters
        ----------
        ptm_dim : dict
            Dictionary containing all of the phantom dimensions that are
            appended as attributes to this class, see class attributes.

        """
        for dimension in ptm_dim.keys():
            setattr(self, dimension, ptm_dim[dimension])

        self.attrs_str = create_attrs_str(attrs_parent=self, object_name="dimensions", indent_level=1)

    def update_attrs_str(self):
        self.attrs_str = create_attrs_str(attrs_parent=self, object_name="dimension", indent_level=1)


class PatientOffset:
    """A class for setting patient - table offset.

    In PyskinDose, the table isocenter is located centered at the head end
    of the support table. The attributes in this class is used to offset the
    patient phantom from this isocenter, in order to get correct patient
    positioning.

    Attributes
    ----------
    d_lat : int
        latertal offset from table isocenter
    d_ver : int
        Vertical offset from table isocenter
    d_lon : int
        longitudianl offset from table isocenter

    Raises
    ------
    NotImplementedError
            Raises error if other units then cm are used.

    """

    def __init__(self, offset: dict):
        """Initialize patient-table offset class.

        Parameters
        ----------
        offset : dict
            offset in cm from the table isocenter in the lateral, vertical and
            longitudinal direction.

        """
        self.d_lat = offset[OFFSET_LATERAL_KEY]
        self.d_ver = offset[OFFSET_VERTICAL_KEY]
        self.d_lon = offset[OFFSET_LONGITUDINAL_KEY]

        self.attrs_str = create_attrs_str(attrs_parent=self, object_name="patient offset", indent_level=1)

    def update_attrs_str(self):
        self.attrs_str = create_attrs_str(attrs_parent=self, object_name="patient offset", indent_level=1)


class Plotsettings:
    """A class for setting plot settings.

    Attributes
    ----------
    interactivity : bool
        Toggle for interactive mode when plotting dosemaps. If True,
        the dosemap will be plotted in a .html file with full interactivity.
        If False, the dosemap will be saved as static images. Static mode is
        provided to enable PySkinDose to run smooth on machines with limited
        RAM.
    dark_mode : bool
        dark mode boolean
    notebook_mode : bool
        Set true if main is called from within a notebook.
        This optimizes plot sizing for notebook output cells.
    plot_dosemap : bool, default is True
        Whether dosemap should be plotted after dose calculation
    max_events_for_patient_inclusion : int
        maximum number of irradiation event for patient inclusion in
        plot_procedure. If the SR file contains more events than this number,
        the patient phantom is not shown in plot_procedure to avoid memory
        error.
    plot_event_index : int
        Index for the event that should be plotted when mode="plot_event" is
        chosen.

    """

    def __init__(self, plt_dict):
        """Initialize plot settings class.

        Parameters
        ----------
        plt_dict : dict
            Dictionary containing all of the plot setting that are
            appended as attributes to this class, see class attributes.
        """
        for key in plt_dict.keys():
            setattr(self, key, plt_dict[key])

        self.attrs_str = create_attrs_str(attrs_parent=self, object_name="plot", indent_level=0)

    def update_attrs_str(self):
        self.attrs_str = create_attrs_str(attrs_parent=self, object_name="plot", indent_level=0)


def create_attrs_str(attrs_parent, object_name, indent_level, indent_size=4, indent_sign=" "):

    attrs_dict = vars(attrs_parent)

    indent_marker_title = (indent_level) * indent_size * indent_sign
    indent_marker_objs = indent_marker_title + indent_size * indent_sign

    attrs_str = indent_marker_title + object_name + "\n"

    for key, val in attrs_dict.items():
        if type(val) in [str, float, bool, int]:
            if not key == "attrs_str":
                attrs_str = attrs_str + indent_marker_objs + str(key) + " : " + str(val) + "\n"
        else:
            attrs_str = attrs_str + "\n" + getattr(attrs_parent, key).attrs_str

    return attrs_str


def initialize_settings(settings: Union[str, dict, PyskindoseSettings]) -> PyskindoseSettings:
    valid_input_settings = settings is not None and isinstance(settings, (str, dict, PyskindoseSettings))

    if not valid_input_settings:
        raise ValueError("Settings must be given as a str or dict")

    if isinstance(settings, PyskindoseSettings):
        return settings

    return PyskindoseSettings(settings=settings)



from pathlib import Path

from setuptools import find_packages, setup

README = (Path(__file__).parent / "README.md").read_text()

setup(
    name="pyskindose",
    version="22.2.1",
    description=(
        "Tools and script for calculating peak skin dose and create dose maps for fluoroscopic exams from RDSR" " data"
    ),
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://dev.azure.com/Sjukhusfysiker/PySkinDose",
    author="Max Hellström",
    author_email="max.hellstrom@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=[
        "pandas",
        "numpy",
        "pydicom>=2.0",
        "numpy-stl",
        "plotly>=4.13.3",
        "scipy",
        "tqdm",
        "psutil",
        "pillow >= 8.1.1",
        "kaleido",
    ],
    include_package_data=True,
    zip_safe=False,
)



from .beam_class import Beam
from .geom_calc import position_patient_phantom_on_table, scale_field_area, fetch_and_append_hvl, check_new_geometry
from .rdsr_parser import rdsr_parser
from .rdsr_normalizer import rdsr_normalizer
from .phantom_class import Phantom
from pyskindose.plotting import plot_geometry
from .analyze_data import analyze_data



import os
import sqlite3

import pandas as pd


def db_connect(db_name: str = "corrections.db"):
    """Set up the database connection with tables needed for PSD calculations.

    Parameters
    ----------
    db_name : str, optional
        The name of/path to the sqlite3 database to connect to
        and if it doesn't exist, create, by default 'corrections.db'

    Returns
    -------
    conn
        connection to database
    cursor
        cursor to database connection

    """
    db_exist = False

    if os.path.exists(db_name):
        db_exist = True

    # Connect to database
    conn = sqlite3.connect(db_name)

    # Create cursor (enables sql commands using the sql method)
    cursor = conn.cursor()

    if not db_exist:

        # load data from CSV files

        # HVL table, simulated with SpekCalc.
        hvl_table = pd.read_csv(os.path.join(os.path.dirname(__file__), "table_data", "HVL_simulated.csv"))

        # Backscatter and mu_en/rho quotients, from Benmanhlouf et al.
        ks_table = pd.read_csv(os.path.join(os.path.dirname(__file__), "table_data", "KS_table_concatenated.csv"))

        # Measured and approximated patient support table transmission.
        tab_pad_table = pd.read_csv(os.path.join(os.path.dirname(__file__), "table_data", "table_transmission.csv"))

        # Table containing lab specific parameters.
        device_info_table = pd.read_csv(os.path.join(os.path.dirname(__file__), "table_data", "device_info.csv"))

        # Upload tables to database

        hvl_table.to_sql("HVL_simulated", conn, if_exists="replace", index=False)

        ks_table.to_sql("KS_table_concatenated", conn, if_exists="replace", index=False)

        tab_pad_table.to_sql("table_transmission", conn, if_exists="replace", index=False)

        device_info_table.to_sql("device_info", conn, if_exists="replace", index=False)

        # Commits the current transactions
        conn.commit()

    return conn, cursor



from pyskindose.constants import (
    KEY_NORMALIZATION_DETECTOR_SIDE_LENGTH,
    KEY_NORMALIZATION_FIELD_SIZE_MODE,
    KEY_NORMALIZATION_MANUFACTURER,
    KEY_NORMALIZATION_MODELS,
    KEY_RDSR_MANUFACTURER,
    KEY_RDSR_MANUFACTURER_MODEL_NAME,
)


class NormalizationSettings:
    """A class to normalize RDSR for PySkinDose compliance.

    Attributes
    ----------
    trans_offset : _TranslationOffset
        See class variables of _TranslationOffset
    trans_dir : _TranslationDirection
        See class variables of _TranslationDirection
    rot_dir : _RotationDirection
        See class variables of _RotationDirection
    field_size_mode : str
        method for calculating field size at image receptor plane.
        Choose either "CFA" (collimated field area) or "ACD" (actual shutter
        distance). For more info, see calculate_field_size in geom_calc.py.
    detector_side_length : str
        side length of active image receptor area in cm.

    """

    def __init__(self, normalization_settings, data_parsed):
        """Initialize class attributes."""
        manufacturer = data_parsed[KEY_RDSR_MANUFACTURER][0]
        model = data_parsed[KEY_RDSR_MANUFACTURER_MODEL_NAME][0]

        # Select correct normalization settings
        for setting in normalization_settings["normalization_settings"]:

            if not (
                manufacturer.lower() == setting[KEY_NORMALIZATION_MANUFACTURER.lower()].lower()
                and model in setting[KEY_NORMALIZATION_MODELS]
            ):
                continue

            self.trans_offset = _TranslationOffset(offset=setting["translation_offset"])

            self.trans_dir = _TranslationDirection(directions=setting["translation_direction"])

            self.rot_dir = _RotationDirection(directions=setting["rotation_direction"])

            self.field_size_mode = setting[KEY_NORMALIZATION_FIELD_SIZE_MODE]
            self.detector_side_length = setting[KEY_NORMALIZATION_DETECTOR_SIDE_LENGTH]
            return

        raise NotImplementedError


class _RotationDirection:
    """Switch pos/neg direction of beam- and table angles.

    Attributes
    ----------
    keys of __init__ parameter directions
        Value of each attribute (i.e. key) is integers of either +1 or -1 to be
        used as a multiplicative correction factor to switch pos/neg direction
        in each angle. PySkinDose default angles are all +1.

    """

    def __init__(self, directions):
        """Initialize class attributes.

        Parameters
        ----------
        directions : dict
            dictionary with keys 'Ap1', 'Ap2', 'Ap3', 'At1', 'At2' and 'At3'.
            Each key contains either '+' or '-'.

        """
        pos_dir = +1
        neg_dir = -1

        for dimension in directions:

            if directions[dimension] == "+":
                setattr(self, dimension, pos_dir)
                continue

            elif directions[dimension] == "-":
                setattr(self, dimension, neg_dir)

            else:
                raise ValueError(f"direction {directions[dimension]} not understood. choose" "either  + or -")

        return


class _TranslationDirection:
    """Switch pos/neg direction of table translations.

    Attributes
    ----------
    keys of __init__ parameter directions
        Value of each attribute (i.e. key) is integers of either +1 or -1 to be
        used as a multiplicative correction factor to switch pos/neg direction
        in each direction. PySkinDose default angles are all +1.

    """

    def __init__(self, directions):
        """Initialize class attributes.

        Parameters
        ----------
        directions : dict
            dictionary with keys 'x', 'y' and 'z'.
            Each key contains either '+' or '-'.

        """
        pos_dir = +1
        neg_dir = -1

        for dimension in directions:

            if directions[dimension] == "+":
                setattr(self, dimension, pos_dir)
                continue

            elif directions[dimension] == "-":
                setattr(self, dimension, neg_dir)

            else:
                raise ValueError(f"direction {directions[dimension]} not understood. choose" "either  + or -")

        return


class _TranslationOffset:
    """Set translation offset of patient support table.

    Use this class to set the translation offset (in cm) between the machine
    origin (the unit that generated the RDSR) and the machine origin of
    PySkinDose (which is located at (x,y,z) = (0, 0, 0)).

    Attributes
    ----------
    keys of __init__ parameter offset
        Value of each attribute (i.e. key) is a float specifying the offset
        in cm.

    """

    def __init__(self, offset):
        """Initialize class attributes.

        Parameters
        ----------
        offset : dict
            dictionary with keys 'x', 'y' and 'z'. Each key contains the
            translation offset (in that direction), specified as a float in cm.

        """
        for dimension in offset:
            setattr(self, dimension, offset[dimension])

        return
